/* This file is Copyright 2001 Level Control Systems.  See the included LICENSE.txt file for details. */

#ifndef MESSAGE_TRANSCEIVER_THREAD_H
#define MESSAGE_TRANSCEIVER_THREAD_H

#include <util/message.h>
#include <util/messenger.h>
#include <util/locker.h>
#include <atheos/semaphore.h>
#include <atheos/threads.h>
#include "iogateway/AbstractMessageIOGateway.h"
#include "atheossupport/ThreadSafeObjectPool.h"

/** What code for the default notification message sent to user code when new PortableMessages have arrived */
#define PORTABLE_MESSAGES_RECEIVED 1349341778 // 'PmRR'

/** Type representing any function that will return a newly allocated AbstractMessageIOGateway object that uses the given PortableDataIO object. */
typedef AbstractMessageIOGateway* (*GatewayFactoryFunc)(PortableDataIO * io);

/**
 *  This is a utility class to help interface multithreaded AtheOS
 *  code more easily to a TCP socket that is sending and receiving
 *  PortableMessages.   Only compiles under AtheOS!
 *  It helps by allowing you to send and receive
 *  PortableMessages without having to worry about calling select(),
 *  or any blocking issues.
 * 
 *  Usage:  
 *          -# Create a new MessageTransceiverThread object.
 *          -# Call StartThread(), StartAcceptThread(), StartServerThread(), or StartConnectThread().
 *          -# Whenever you wish to send a PortableMessage over the TCP socket,
 *             simply call SendOutgoingMessage().  
 *          -# Whenever a PortableMessage is received, a PORTABLE_MESSAGES_RECEIVED
 *             Message will be sent to the Messenger you specified in StartThread().
 *             When you get this, you can respond by calling GetNextIncomingMessage()
 *             on this object to retrieve the PortableMessage.
 *          -# When you are done, just delete this object normally.  If you wish,
 *             you can call WaitForAllMessagesToBeSent() first, in order to ensure
 *             that all pending outbound messages get delivered.
 *          -# That's it!
 */
class MessageTransceiverThread
{
public:
   /** Creates a receiver thread without a reply target.  You should
     * call SetReplyTarget() before calling StartThread()...
     * @param threadPri the AtheOS thread priority the TCP thread should run at.  Defaults to B_NORMAL_PRIORITY.
     */
   MessageTransceiverThread(int32 threadPri = NORMAL_PRIORITY);

   /** Creates a receiver thread with a reply target.
    *  @param replyTarget specifies where you want the PORTABLE_MESSAGES_RECEIVED (and other messages) sent to.
    *  @param threadPri is the AtheOS thread priority the TCP thread should run at.  Defaults to NORMAL_PRIORITY.
    */
   MessageTransceiverThread(const os::Messenger & replyTarget, int32 threadPri = NORMAL_PRIORITY);

   /** 
    *  Destructor.  Safe to call at any time. 
    */
   virtual ~MessageTransceiverThread();

   /** Creates a TCP send/receive thread that connects to the given hostname and port.
    *  @param hostname is the Internet hostname of the computer to connect to.
    *  @param port is the port number to connect on.
    *  @param optConnectMessage if specified, will be sent via your Messenger when the connection 
    *                           is complete.  (A copy of this message is made).
    *  @param optDisconnectMessage if specified, will be sent via your Messenger when the connection 
    *                              fails or is broken. (A copy of this message is made).
    *  @return B_NO_ERROR if the thread was started successfully, or B_ERROR if it failed.
    *          (note that this method occasionally fails due to a bug in the net_server; when
    *           that happens, the only solution is to have the user restart networking.  BONE fixes the bug.)
    */
   status_t StartConnectThread(const char * hostname, int port, 
                               const os::Message * optConnectMessage = NULL, const os::Message * optDisconnectMessage = NULL);

   /**
    * Creates a TCP send/receive thread that waits for connections on the given port.
    *  @param port the port to listen for incoming connections on.  If 0, then the system 
    *              will choose a port to accept on.  The port number it chooses can be discovered by calling GetPort().
    *  @param acceptMultiple if true, will allow subsequent connections to be accepted after the first one is closed.
    *                        In no case will there be more than one connection active at a time, however.
    *  @param optConnectMessage if specified, will be sent via your Messenger when the connection 
    *                           is complete.  (A copy of this message is made).
    *  @param optDisconnectMessage if specified, will be sent via your Messenger when the connection 
    *                              fails or is broken. (A copy of this message is made).
    *  @return B_NO_ERROR if the thread was started successfully, or B_ERROR if it failed.
    *          (note that this method occasionally fails due to a bug in the net_server; when
    *           that happens, the only solution is to have the user restart networking.  BONE fixes the bug.)
    */
   status_t StartAcceptThread(int port, bool acceptMultiple,
                              const os::Message * optConnectMessage = NULL, const os::Message * optDisconnectMessage = NULL);

   /**
    * Creates a TCP accept thread that waits for connections on the given port.
    * The difference between calling this call and StartAcceptThread() lies in how the
    * incoming connection is handled.  With StartAcceptThread(), the incoming connection
    * becomes the 'current connection' for this object, and PortableMessages are sent and
    * received as usual.  With this method, on the other hand, an incoming connection is
    * accepted and placed into an incoming-connections queue, and the (connectMessage) you
    * specify is sent to your handler.  Your handler should then call GetNextIncomingConnection() 
    * on this object to retrieve the socket for the new connection, which can be handled with
    * a new MessageTransceiverThread or however you like. 
    *  @param port the port to listen for incoming messages on.  If set to 0, the system will 
    *         choose a port to accept on.  The port number it chooses can be discovered by calling GetPort().
    *  @param acceptMultiple if false, then only the first incoming connection will ever be accepted.
    *  @return B_NO_ERROR if the thread was started successfully, or B_ERROR if it failed.
    *          (note that this method occasionally fails due to a bug in the net_server; when
    *           that happens, the only solution is to have the user restart networking.  BONE fixes the bug.)
    */
   status_t StartServerThread(int port, const os::Message & connectMessage, bool acceptMultiple=true);

   /**
    * Creates and activates the TCP send/receive background thread.
    *  @param gateway is the AbstractMessageIOGateway the thread should use to send and receive data.
    *                 In general, a AbstractMessageIOGateway object with a non-blocking TCPSocketDataIO
    *                 that uses your socket is the way to go here.
    *  @param optConnectMessage if specified, will be sent via your Messenger when the connection 
    *                      is complete.  (A copy of this message is made).
    *  @param optDisconnectMessage if specified, will be sent via your Messenger when the connection 
    *                         fails or is broken. (A copy of this message is made).
    *  @return B_NO_ERROR if the thread was started successfully, or B_ERROR if it failed.
    */
   status_t StartThread(AbstractMessageIOGateway * gateway, int datafd, const os::Message * optDisconnectMessage = NULL);
   
   /**
    * Call this to retrieve the next message from the incoming messages queue.
    * Otherwise returns B_ERROR.
    *  @param setRef On success, this reference will contain the next received PortableMessage for you to look at.
    *  @return B_NO_ERROR on success, or B_ERROR if there are no more PortableMessages in the incoming messages queue.
    */
   status_t GetNextIncomingMessage(PortableMessageRef & setRef);

   /**
    * Enqueues a PortableMessage for immediate output over the TCP connection.
    * @param msgRef a reference to the PortableMessage to send out.
    * @return B_NO_ERROR on success, B_ERROR if no more messages can be sent (e.g. connection is broken)
    * NOTE:  If you want to use object pools with your (msgRef), make sure that the
    *        pools you use are ThreadSafeObjectPools objects, or you will suffer from
    *        nasty race conditions.  The GetThreadSafeMessagePool() and GetThreadSafeMessageRefPool()
    *        methods in this class return appropriate pool objects to use here.
    */
   status_t SendOutgoingMessage(const PortableMessageRef & msgRef);

   /**
    * Blocks until all messages that were in the outgoing queue have been sent.
    * @return B_NO_ERROR on success, B_ERROR if there was an error.
    */
   void WaitForAllMessagesToBeSent();

   /**
    * Returns the port number that the accept thread is accepting messages on. 
    * This number is only meaningful if StartAcceptThread() or StartServerThread() was called.
    */
   int GetPort() const;

   /**
    * Sets or resets the reply target for this thread
    * @param target the new target that this MessageTransceiverThread should send its Message to.
    */
   void SetReplyTarget(const os::Messenger & target);

   /**
    * If you wish to use a type of AbstractMessageIOGateway other than the default type,
    * you can set a callback to your own factory function here.  The factory function should
    * return a new AbstractMessageIOGateway object (or subclass thereof), or NULL on failure.
    * The returned object becomes property of this MessageTransceiverThread object.
    * @param func the new factory function this MessageTransceiverThread should use to create its AbstractMessageIOGateway.
    */
   void SetGatewayFactoryFunc(GatewayFactoryFunc func);

   /**
    * Call this if you wish to receive a Message every time the outgoing message queue empties.  
    * @param msg points to the Message to send, or be NULL to turn off empty-queue-messaging.
    * Caller retains ownership of the Message (a copy of (*msg) will be made)
    */
   void SetOutgoingQueueDrainedMessage(const os::Message * msg);

   /** 
    * Call this method if you wish to customize the notification Message that this
    * MessageTransceiverThread sends to tell your code that new PortableMessages have
    * arrived.  (By default, this message is empty and has a 'what' code of 
    * PORTABLE_MESSAGES_RECEIVED, but you might want to change that if, for example,
    * you have several MessageTransceiverThreads sending to the same target, and it
    * needs to distinguish which MessageTransceiverThread the Messages came from)
    * @param notifyMsg The new notification Message to use.
    */
   void SetNotificationMessage(const os::Message & notifyMsg) {_notifyMsg = notifyMsg;}

   /**
    * This call is only useful for MessageTransceiverThreads that are running in server mode
    * (i.e. that you have called StartServerThread() on).  It will retrieve the next queued
    * incoming TCP connection socket and copy it into (returnSocketFD).  If a socket is 
    * returned to you, your code becomes the owner of the socket.
    * @param returnSocketFD will be set to the file descriptor of the new socket on success.
    * @return B_NO_ERROR on success, or B_ERROR if there is no incoming connection ready.
    */
   status_t GetNextIncomingConnection(int & returnSocketFD);

   /**
    * Returns a pointer to a singleton thread-safe message pool.
    * Equivalent to ::GetMessagePool(), but for use in multithreaded AtheOS environments.
    */
   static ThreadSafeObjectPool<PortableMessage> * GetMessagePool() {return &_msgPool;}

   /**
    * Returns a pointer to a singleton thread-safe message ref pool.
    * Equivalent to ::GetMessageRefPool(), but for use in multithreaded AtheOS environments.
    */
   static ThreadSafeObjectPool<PortableMessageRef::RefCountMem> * GetMessageRefPool() {return &_msgRefPool;}

private:
   AbstractMessageIOGateway * CreateIOGateway(int socketfd) const;

   status_t StartThread(void * const func, const char * threadName);
   void NotifyThread();
   status_t CreateNotificationMechanism();
   static void TCPThreadStub(void * d) {((MessageTransceiverThread *)d)->TCPThreadLoop();}
   static void ConnectThreadStub(void * d) {((MessageTransceiverThread *)d)->ConnectThreadLoop();}
   void AcquireMessagesPendingLock();
   void ReleaseMessagesPendingLock();
   void TCPThreadLoop();
   void ConnectThreadLoop();

   volatile thread_id _tcpThread;
   os::Messenger _replyTarget;
   os::Message _notifyMsg;
   os::Locker _lock;   

   // _gateway and _datafd are linked.  If _gateway if non-NULL, _datafd must be a valid fd.
   // And if _gateway is NULL, _datafd must be < 0.
   AbstractMessageIOGateway * _gateway;
   int _datafd;    // the fd we actually transfer messages over

   int _acceptfd;  // the fd we wait for connections on (or -1 if we aren't doing that)

   int _notifyfd;  // the fd we use to wake up our thread (hacksville!)
   int _waitfd;    // the fd the thread uses to be woken up (hack hack!)

   PortableString _connectTo;  // used by StartConnectThread()
   int _port;
   int32 _threadPri;
 
   bool _isConnecting;
   bool _isServer;      // true iff StartServerThread() was called

   os::Message _onConnect;
   bool _sendOnConnect;

   os::Message _onDisconnect;
   bool _sendOnDisconnect;

   os::Message _onEmpty;
   bool _sendOnEmpty;

   bool _acceptMultiple;

   sem_id _messagesPendingSemaphore;
   atomic_t _messagesPendingBenaphore;

   GatewayFactoryFunc _userGatewayFunc;

   PortableQueue<int> _incomingConnections;  // used when in server mode

   static ThreadSafeObjectPool<PortableMessage> _msgPool;
   static ThreadSafeObjectPool<PortableMessageRef::RefCountMem> _msgRefPool;
};

#endif
