#include <stdio.h>

#include <app/Looper.h>
#include "reflector/StorageReflectConstants.h"
#include "besupport/MessageTransceiverThread.h"

// This BLooper subclass is used to process the incoming
// BMessages that our MessageTransceiverThread sends to it
class MyLooper : public BLooper
{
public:
   MyLooper(const char * userName, const char * serverName)
   {
      _userName = userName;  // save this for the ReadyToRun() call
      _serverName = serverName;  // "beshare.bentonrea.com" or whatnot
      _mtt = NULL; // just in case
   }

   ~MyLooper()
   {
      delete _mtt;  // shut down and get rid of the MessageTransceiverThread
   }
   
   virtual thread_id Run()
   {
      thread_id tid = BLooper::Run();
      if (tid >= 0)
      {
         BMessage connectMsg('conn');     // message we want to see when the connection is made
         BMessage disconnectMsg('disc');  // message we want to see when the connection is broken

         // Startup!  Create our MessageTransceiverThread... that is our
         // interface to MUSCLE and to the Internet.
         _mtt = new MessageTransceiverThread(BMessenger(this));  // have it send replies back to me
         if (_mtt->StartConnectThread(_serverName.Cstr(), 2960, &connectMsg, &disconnectMsg) == B_NO_ERROR)
         {
            printf("Now connecting to the MUSCLE server...\n");
         }
         else
         {
            printf("Aggh!  StartConnectThread() failed... net_server needs restarting???\n");
            PostMessage(B_QUIT_REQUESTED);
         }
      }
      return tid;
   }

   virtual void MessageReceived(BMessage * msg)
   {
      printf("\n\n");
      switch(msg->what)
      {
         case 'conn':
         {
            printf("MUSCLE says the TCP connection to the MUSCLE server has been completed/connected!!\n");

            {
               // Now that we are connected, upload a node to the server to tell everyone
               // that we are a clientserverdemo client (and not, say, a BeShare client)
               PortableMessage * data = new PortableMessage;  // this could be any message, but we'll leave it empty
               PortableMessageRef dataMsgRef(data, NULL, NULL);

               // This is the actual command to the server to set an ID node for us
               PortableMessage * setCommand = new PortableMessage(PR_COMMAND_SETDATA);
               setCommand->AddMessage("clientserverdemo", dataMsgRef);
               PortableMessageRef setCommandRef(setCommand, NULL, NULL);
               _mtt->SendOutgoingMessage(setCommandRef);
            }

            {
               // Lastly, let's "subscribe" to the clientserverdemo node.
               // This way, the muscle daemon will tell us whenever another
               // clientserverdemo node is added or removed from his database;
               // that lets us know if another clientserverdemo program has
               // logged in or logged out.
               PortableMessage * subscribeMessage = new PortableMessage(PR_COMMAND_SETPARAMETERS);
               subscribeMessage->AddBool("SUBSCRIBE:clientserverdemo", true);
               PortableMessageRef subscribeRef(subscribeMessage, NULL, NULL);
               _mtt->SendOutgoingMessage(subscribeRef);
            }
         }
         break;

         case 'disc':
            printf("MUSCLE says the TCP connection  to the MUSCLE serverhas been broken/disconnected!!\n");
         break;

         case PORTABLE_MESSAGES_RECEIVED:
         {
            printf("MUSCLE says it got some PortableMessages from the TCP connection!  They are:\n");
            PortableMessageRef msgRef;
            while(_mtt->GetNextIncomingMessage(msgRef) == B_NO_ERROR)
            {
               PortableMessage * pmsg = msgRef.GetItemPointer();
               pmsg->PrintToStream();
            }
         }
         break;

         default:
            BLooper::MessageReceived(msg);
         break;
      }
   }

   MessageTransceiverThread * GetMessageTransceiverThread() {return _mtt;}

private:
   PortableString _userName;
   PortableString _serverName;
   MessageTransceiverThread * _mtt;
};

void main(int argc, char ** argv)
{
   if (argc < 2)
   {
      printf("Usage:  clientserver username [servername]\n");
      exit(5);
   }

   // Create and start up our incoming-messages-processing looper
   MyLooper * myLooper = new MyLooper(argv[1], (argc > 2) ? argv[2] : "beshare.bentonrea.com");
   myLooper->Run();
 
   // Now we will just watch stdin for user input.  A program with
   // a GUI would probably create a BApplication and call Run() on
   // that instead, but we can't be bothered with a GUI in this example.
   while(1)
   {
      char text[1000];
      gets(text);
      printf("You typed: [%s]\n",text);
      if (strcmp(text, "quit") == 0)
      {
         printf("Quitting!\n");
         break;
      }
      else
      {
         PortableMessage * msg = new PortableMessage('chat');
         msg->AddString("text", text);
         msg->AddString("from", argv[1]);  // so they can know who said it

         // Add this key string to tell the muscle daemon to forward this message 
         // to other clientserverdemo clients (and not to e.g. BeShare clients or 
         // battleship clients or other clients that don't care)
         msg->AddString(PR_NAME_KEYS, "clientserverdemo"); 

         printf("Sending a PortableMessage out to the MessageTranceiverThread... the message is:\n");
         msg->PrintToStream();

         PortableMessageRef msgRef(msg, NULL, NULL);
         myLooper->GetMessageTransceiverThread()->SendOutgoingMessage(msgRef);
      }
   }

   if (myLooper->Lock()) myLooper->Quit();  // shut down the looper

   printf("Bye bye!\n"); 
   return 0;
}
