/* This file is Copyright 2001 Level Control Systems.  See the included LICENSE.txt file for details. */

/*******************************************************************************
/
/   File:      Rect.h
/
/   Description:    Portable version of Be's BRect class
/
*******************************************************************************/

#ifndef PORTABLE_RECT_H
#define PORTABLE_RECT_H

#include <math.h>
#include "support/PortablePoint.h"

/*-----------------------------------------------------------------------*/
/*----- PortableRect class ----------------------------------------------*/

/** A portable version of Be's BRect class. */
class PortableRect : public PortableFlattenable
{
public:
   /** Exposed member variable representing the position of the left edge of the rectangle. */
   float left;

   /** Exposed member variable representing the position of the top edge of the rectangle. */
   float top;

   /** Exposed member variable representing the position of the right edge of the rectangle. */
   float right;

   /** Exposed member variable representing the position of the bottom edge of the rectangle. */
   float bottom;

   /** Default Constructor.  
     * Creates a rectangle with upper left point (0,0), and lower right point (-1,-1).
     * Note that this rectangle has a negative area!  (that is, it's imaginary)
     */ 
   PortableRect() {left = top = 0.0f;  right = bottom = -1.0f;}

   /** Constructor where you specify the left, top, right, and bottom coordinates */
   PortableRect(float l, float t, float r, float b) {Set(l,t,r,b);}

   /** Copy constructor */
   PortableRect(const PortableRect & rhs) : PortableFlattenable() {*this = rhs;}

   /** Constructor where you specify the leftTop point and the rightBottom point. */
   PortableRect(PortablePoint leftTop, PortablePoint rightBottom) {Set(leftTop.x, leftTop.y, rightBottom.x, rightBottom.y);}

   /** Destructor */
   virtual ~PortableRect() {/* empty */}

   /** Set a new position for the rectangle. */
   void Set(float l, float t, float r, float b)
   {
      left   = l;
      top    = t;
      right  = r;
      bottom = b;
   }

   /** Print the rectangle's current state to stdout */
   void PrintToStream() const
   {
      printf("PortableRect: leftTop=(%f,%f) rightBottom=(%f,%f)\n", left, top, right, bottom);
   }

   /** Returns the left top corner of the rectangle. */
   PortablePoint LeftTop() const {return PortablePoint(left, top);}

   /** Returns the right bottom corner of the rectangle. */
   PortablePoint RightBottom() const {return PortablePoint(right, bottom);}

   /** Returns the left bottom corner of the rectangle. */
   PortablePoint LeftBottom() const {return PortablePoint(left, bottom);}

   /** Returns the right top corner of the rectangle. */
   PortablePoint RightTop() const {return PortablePoint(right, top);}

   /** Set the left top corner of the rectangle. */
   void SetLeftTop(const PortablePoint p) {left = p.x; top = p.y;}

   /** Set the right bottom corner of the rectangle. */
   void SetRightBottom(const PortablePoint p) {right = p.x; bottom = p.y;}

   /** Set the left bottom corner of the rectangle. */
   void SetLeftBottom(const PortablePoint p) {left = p.x; bottom = p.y;}

   /** Set the right top corner of the rectangle. */
   void SetRightTop(const PortablePoint p) {right = p.x; top = p.y;}

   /** Makes the rectangle smaller by the amount specified in both the x and y dimensions */
   void InsetBy(PortablePoint p) {InsetBy(p.x, p.y);}

   /** Makes the rectangle smaller by the amount specified in both the x and y dimensions */
   void InsetBy(float dx, float dy) {left += dx; top += dy; right -= dx; bottom -= dy;}

   /** Translates the rectangle by the amount specified in both the x and y dimensions */
   void OffsetBy(PortablePoint p) {OffsetBy(p.x, p.y);}

   /** Translates the rectangle by the amount specified in both the x and y dimensions */
   void OffsetBy(float dx, float dy) {left += dx; top += dy; right += dx; bottom += dy;}

   /** Translates the rectangle so that its top left corner is at the point specified. */
   void OffsetTo(PortablePoint p) {OffsetTo(p.x, p.y);}

   /** Translates the rectangle so that its top left corner is at the point specified. */
   void OffsetTo(float x, float y) {right = x + Width(); bottom = y + Height(); left = x; top = y;}

   /** Comparison Operator.  Returns true iff (r)'s dimensions are exactly the same as this rectangle's. */
   bool operator==(PortableRect r) const {return (left == r.left)&&(top == r.top)&&(right == r.right)&&(bottom == r.bottom);}

   /** Comparison Operator.  Returns true iff (r)'s dimensions aren't exactly the same as this rectangle's. */
   bool operator!=(PortableRect r) const {return !(r == *this);}

   /** Returns a rectangle whose area is the intersecting subset of this rectangle's and (r)'s */
   PortableRect operator&(PortableRect r) const 
   {
      PortableRect ret(*this);
      if (ret.left   < r.left)   ret.left   = r.left;
      if (ret.right  > r.right)  ret.right  = r.right;
      if (ret.top    < r.top)    ret.top    = r.top;
      if (ret.bottom > r.bottom) ret.bottom = r.bottom;
      return ret;
   }

   /** Returns a rectangle whose area is a superset of the union of this rectangle's and (r)'s */
   PortableRect operator|(PortableRect r) const 
   {
      PortableRect ret(*this);
      if (r.left   < ret.left)   ret.left   = r.left;
      if (r.right  > ret.right)  ret.right  = r.right;
      if (r.top    < ret.top)    ret.top    = r.top;
      if (r.bottom > ret.bottom) ret.bottom = r.bottom;
      return ret;
   }

   /** Returns true iff this rectangle and (r) overlap in space. */
   bool Intersects(PortableRect r) const {return (r&(*this)).IsValid();}

   /** Returns true iff this rectangle's area is non imaginary (i.e. Width() and Height()) are both non-negative) */
   bool IsValid() const {return ((Width() >= 0.0f)&&(Height() >= 0.0f));}

   /** Returns the width of this rectangle. */
   float Width() const {return right - left;}

   /** Returns the width of this rectangle as an integer. */
   int32 IntegerWidth() const {return (int32)ceil(Width());}

   /** Returns the height of this rectangle. */
   float Height() const {return bottom - top;}

   /** Returns the height of this rectangle as an integer. */
   int32 IntegerHeight() const {return (int32)ceil(Height());}

   /** Returns true iff this rectangle contains the specified point. */
   bool Contains(PortablePoint p) const
   {
      return ((p.x >= left)&&(p.x <= right)&&(p.y >= top)&&(p.y <= bottom));
   }

   /** Returns true iff this rectangle fully the specified rectangle. */
   bool Contains(PortableRect p) const
   {
      return ((Contains(p.LeftTop()))&&(Contains(p.RightTop()))&&
              (Contains(p.LeftBottom()))&&(Contains(p.RightBottom())));
   }

   /** Part of the PortableFlattenable API:  Returns true. */
   virtual bool IsFixedSize() const {return true;}

   /** Part of the PortableFlattenable API:  Returns B_RECT_TYPE. */
   virtual type_code TypeCode() const {return B_RECT_TYPE;}

   /** Part of the PortableFlattenable API:  Returns 4*sizeof(float). */
   virtual size_t FlattenedSize() const {return 4 * sizeof(float);}

   /** Flattens this rectangle into an endian-neutral byte buffer.
    *  @param buffer Points to the byte buffer to write into.  There must be at least FlattenedSize() bytes there. 
    */
   virtual void Flatten(uint8 * buffer) const
   {
      float * buf = (float *) buffer;
      buf[0] = B_HOST_TO_LENDIAN_FLOAT(left);
      buf[1] = B_HOST_TO_LENDIAN_FLOAT(top);
      buf[2] = B_HOST_TO_LENDIAN_FLOAT(right);
      buf[3] = B_HOST_TO_LENDIAN_FLOAT(bottom);
   }

   /** Unflattens this rectangle from an endian-neutral byte buffer.
    *  @param buffer Points to the byte buffer to read data from.
    *  @param size The number of bytes available in (buffer).  There should be at least FlattenedSize() bytes there.
    *  @return B_NO_ERROR on success, or B_ERROR if (buffer) was too small.
    */
   virtual status_t Unflatten(const uint8 * buffer, size_t size)
   {
      if (size >= FlattenedSize())
      {
         float * buf = (float *) buffer;
         left   = B_LENDIAN_TO_HOST_FLOAT(buf[0]);
         top    = B_LENDIAN_TO_HOST_FLOAT(buf[1]);
         right  = B_LENDIAN_TO_HOST_FLOAT(buf[2]);
         bottom = B_LENDIAN_TO_HOST_FLOAT(buf[3]);
         return B_NO_ERROR;
      }
      else return B_ERROR;
   }
};

#endif 
