#include <stdio.h>
#include "app/Application.h"      // Be's BApplication class, necessary for any GUI stuff
#include "interface/Window.h"     // Be's BWindow class
#include "interface/View.h"       // Be's BView class
#include "interface/StringView.h" // Be's BStringView class, for showing a text label
#include "interface/Button.h"     // Be's BButton class
#include "interface/Rect.h"       // Be's BRect class

const uint32 COMMAND_INCREMENT_COUNTER = 'Incr';

/* My custom BWindow subclass... */
class MessageDemoWindow : public BWindow
{
public:
   MessageDemoWindow(BRect rect);
   
   virtual void MessageReceived(BMessage * msg);
   virtual bool QuitRequested();

   void SetButtonMessageTarget(BMessenger target);

private:
   uint32 _counter;            // counts how many COMMAND_INCREMENT_COUNTER messages we've received so far
   BButton * _button;          // button for the user to press to send a message out
   BView * _background;        // our colored background view
   BStringView * _countView;   // text label / displays number of messages received
   BMessenger _messageTarget;  // where to send a message when the user clicks our button
};

MessageDemoWindow :: MessageDemoWindow(BRect rect)
   : BWindow(rect, "MessageDemoWindow!", B_TITLED_WINDOW, 0), _counter(0)
{
   // Fill this window with a nice light green background BView...
   _background = new BView(Bounds(), NULL, B_FOLLOW_ALL_SIDES, 0);
   _background->SetViewColor(100,255,0);
   AddChild(_background);

   // And put a nice BButton inside the view.  This button will send a COMMAND_INCREMENT_COUNTER
   // message to its target (which we will specify later) whenever it is clicked.
   BRect buttonRect = _background->Bounds();
   buttonRect.InsetBy(5,10);  // we'll make the button a bit smaller than the view
   buttonRect.bottom -= 20.0f;
   _button = new BButton(buttonRect, NULL, "Send Message!", new BMessage(COMMAND_INCREMENT_COUNTER), B_FOLLOW_ALL_SIDES, B_WILL_DRAW|B_FULL_UPDATE_ON_RESIZE);
   _button->SetTarget(this);  // so that when the button is pressed, its BMessage is sent to this window.
   _background->AddChild(_button);

   // Make a little text display to show how many messages we have received
   BRect displayRect(_background->Bounds());
   displayRect.bottom -= 5.0f;
   displayRect.top = displayRect.bottom - 20.0f;
   _countView = new BStringView(displayRect, NULL, "(No increment messages received yet)", B_FOLLOW_BOTTOM | B_FOLLOW_LEFT_RIGHT);
   _countView->SetAlignment(B_ALIGN_CENTER);  // center the text so it is prettier
   _background->AddChild(_countView);
}

void
MessageDemoWindow :: SetButtonMessageTarget(BMessenger target)
{
   _button->SetTarget(target);  // tell the button where to send its message when pressed
}


// Called whenever our window gets a BMessage.  
void MessageDemoWindow :: MessageReceived(BMessage * msg)
{
   switch(msg->what)
   {
      case COMMAND_INCREMENT_COUNTER:
      {
         // We got a message telling us to increment our counter!  
         _counter++;
         char buf[128];
         sprintf(buf, "Total Messages Received:  %lu", _counter);
         _countView->SetText(buf);

         {
            // Also, turn our background yellow to indicate we got something
            // Is there a better way to do this?

            const rgb_color yellow = {255, 255, 0, 255};
            _background->SetViewColor(yellow);
            _background->Invalidate();

            _button->SetViewColor(yellow);
            _button->SetLowColor(yellow);
            _button->Invalidate();

            _countView->SetViewColor(yellow);
            _countView->SetLowColor(yellow);
            _countView->Invalidate();
         }
      }
      break;

      default:
         BWindow :: MessageReceived(msg);  // all other types go to the parent class's method.
      break;
   }
}

// Called when the user clicks on the window's close box
bool MessageDemoWindow :: QuitRequested()
{
   // Note that there may be a race condition here... can you see it?
   if (be_app->CountWindows() == 1) be_app->PostMessage(B_QUIT_REQUESTED);  // last window?  Tell the application to quit too

   return true;  // and indicate that it's okay for us to close
}

// Demonstrates a BWindow with a BView and a BButton.  Compile like this:
// gcc window.cpp -lbe
int main(int argc, char ** argv)
{
   BApplication app("application/x-vnd.Sugoi-sillybuttondemo");  // gotta have this!

   // Create two windows...
   MessageDemoWindow * window1 = new MessageDemoWindow(BRect(50,50,275,125));
   MessageDemoWindow * window2 = new MessageDemoWindow(BRect(50,200,275,275));

   // Tell them to send their messages to each other when their button is clicked
   window1->SetButtonMessageTarget(window2);
   window2->SetButtonMessageTarget(window1);

   window1->Show();  // starts window1's thread and displays the window
   window2->Show();  // starts window2's thread and displays the window

   app.Run();    // this call won't return until the application is ready to quit

   printf("bye!\n");

   return 0;
}
